/*
 * Copyright (c) 1996 Shunsuke Akiyama.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by Shunsuke Akiyama.
 * 4. Neither the name of the author nor the names of any co-contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY Shunsuke Akiyama AND CONTRIBUTORS ``AS IS''
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL Shunsuke Akiyama OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 *	$Id: odcontrol.c,v 1.5 1996/11/17 13:49:35 shun Exp $
 */

/*
 * odcontrol - SCSI optical disk drive control utility
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <sys/scsiio.h>
#include <sys/file.h>
#include <scsi.h>
#include <sys/stat.h>
#include <sys/cdio.h>

/*
 * default timeout
 */

#define DEFAULT_TIMEOUT		10000	/* normal timeout value, 10000 mS */
#define DEFAULT_TIMEOUT_LONG	3600000	/* long timeout value, 3600 Sec */

/*
 * SCSI commands
 */

#define SCSI_TEST_UNIT_READY	0x00
#define SCSI_FORMAT_UNIT	0x04
#define SCSI_REASSIGN_BLOCKS	0x07
#define SCSI_MODE_SELECT	0x15
#define SCSI_MODE_SENSE		0x1a
#define SCSI_READ_CAPACITY	0x25
#define SCSI_VERIFY		0x2f

/*
 * SCSI sense key code
 */

#define SCSI_SENSE_KEY		0x0f

#define SCSI_NO_SENSE		0
#define SCSI_RECOVERD_ERROR	1
#define SCSI_NOT_READY		2
#define SCSI_MEDIUM_ERROR	3
#define SCSI_HARDWARE_ERROR	4
#define SCSI_ILLEGAL_REQUEST	5
#define SCSI_UNIT_ATTENTION	6
#define SCSI_DATA_PROTECT	7
#define SCSI_BLANK_CHECK	8
#define SCSI_VENDOR_UNIQUE	9
#define SCSI_COPY_ABORTED	10
#define SCSI_ABORTED_COMMAND	11
#define SCSI_EQUAL		12
#define SCSI_VOLUME_OVERFLOW	13
#define SCSI_MISCOMPARE		14
#define SCSI_RESERVED		15

/*
 * SCSI sense data
 */

struct sense_data {
	u_char	err_code;
	u_char	seg_no;
	u_char	sense_key;
	u_char	info_3;		/* MSB */
	u_char	info_2;
	u_char	info_1;
	u_char	info_0;		/* LSB */
	u_char	add_sense_len;
	u_char	c_info_3;	/* MSB */
	u_char	c_info_2;
	u_char	c_info_1;
	u_char	c_info_0;	/* LSB */
	u_char	asc;
	u_char	ascq;
	u_char	fru;
	u_char	sk_info_2;	/* MSB */
	u_char	sk_info_1;
	u_char	sk_info_0;	/* LSB */
};

/*
 * SCSI TEST UNIT READY command
 */

struct test_unit_ready {
	u_char	op_code;
	u_char	lun;
	u_char	reserved1;
	u_char	reserved2;
	u_char	reserved3;
	u_char	control;
};

/*
 * SCSI FORMAT UNIT command
 */

struct format_unit {
	u_char	op_code;
	u_char	lun_flag;
	u_char	vendor_uniq;
	u_char	interleave_1;
	u_char	interleave_0;
	u_char	control;
};

/*
 * SCSI REASSIGN BLOCKS command and parameter
 */

/* REASSIGN BLOCKS command */
struct reassign_blocks {
	u_char	op_code;
	u_char	lun;
	u_char	reserve1;
	u_char	reserve2;
	u_char	reserve3;
	u_char	control;
};

/* defect list */
struct defect_list {
	/* defect list header */
	u_char	reserve1;
	u_char	reserve2;
	u_char	list_len_1;	/* MSB */
	u_char	list_len_0;	/* LSB */
	/* defect descriptor */
	u_char	block_addr_3;	/* MSB */
	u_char	block_addr_2;
	u_char	block_addr_1;
	u_char	block_addr_0;	/* LSB */
};

/*
 * SCSI MODE SELECT/SENSE command and parameter
 */

#define SCSI_MODE_PAGE8_RCD	1	/* read cache disable */
#define SCSI_MODE_PAGE8_WCE	4	/* write cache enable */

/* MODE SELECT command */
struct mode_select {
	u_char	op_code;
	u_char	lun_flag;
	u_char	reserved1;
	u_char	reserved2;
	u_char	plist_len;
	u_char	control;
};

/* MODE SENSE command */
struct mode_sense {
	u_char	op_code;
	u_char	lun_flag;
	u_char	page;
	u_char	reserved;
	u_char	alloc_len;
	u_char	control;
};

/* mode parameter header */
struct mode_parm_hdr {
	u_char	mode_parm_len;
	u_char	media_type;
	u_char	dev_parm;
	u_char	blk_desc_len;
};

/* cache control parameter */
struct cache_ctl {
	u_char	page;
	u_char	page_len;
	u_char	cache_mode;
	u_char	priority;
	u_char	prefetch_determent_1;	/* MSB */
	u_char	prefetch_determent_0;	/* LSB */
	u_char	min_prefetch_1;		/* MSB */
	u_char	min_prefetch_0;		/* LSB */
	u_char	max_prefetch_1;		/* MSB */
	u_char	max_prefetch_0;		/* LSB */
	u_char	prefetch_limit_1;	/* MSB */
	u_char	prefetch_limit_0;	/* LSB */
};

/*
 * SCSI READ CAPACITY command and parameter
 */

/* READ CAPACITY command */
struct read_capacity {
	u_char	op_code;
	u_char	lun_flag;
	u_char	block_addr_3;	/* MSB */
	u_char	block_addr_2;
	u_char	block_addr_1;
	u_char	block_addr_0;	/* LSB */
	u_char	reserved1;
	u_char	reserved2;
	u_char	reserved3;
	u_char	control;
};

/* capacity data */
struct read_capacity_data {
	u_char	block_addr_3;	/* MSB */
	u_char	block_addr_2;
	u_char	block_addr_1;
	u_char	block_addr_0;	/* LSB */
	u_char	block_len_3;	/* MSB */
	u_char	block_len_2;
	u_char	block_len_1;
	u_char	block_len_0;	/* LSB */
};

/*
 * SCSI VERIFY command
 */

struct verify {
	u_char	op_code;
	u_char	lun_flag;
	u_char	block_addr_3;	/* MSB */
	u_char	block_addr_2;
	u_char	block_addr_1;
	u_char	block_addr_0;	/* LSB */
	u_char	reserved;
	u_char	block_len_1;	/* MSB */
	u_char	block_len_0;	/* LSB */
	u_char	control;
};

#define MAX_VERIFY_LENGTH	0x800

/*
 * Error codes
 */

#define ERR_BADBLOCK	-1
#define ERR_NONE	0
#define ERR_COMMAND	1
#define ERR_SYSCALL	2
#define ERR_SCSI	3

/*
 * command table
 */

struct cmd_tab {
	char	*c_name;
	char	*help;
	int	(*func)();
	char	*def_dev;
};

/*
 * cache mode control argument
 */

char *cache_args[] = { "off", "write", "read", "on", NULL };

#define C_CACHE_OFF	0
#define C_CACHE_WR	1
#define C_CACHE_RD	2
#define C_CACHE_ON	3

/*
 * command table
 */

int exec_cache(), exec_eject(), exec_format(), exec_verify();

struct cmd_tab commands[] = {
	{
		"cache",
		"[on|read|write|off] (drive cache control)",
		exec_cache,
		"/dev/rod0.ctl"
	},
	{
		"eject",
		"(eject a media)",
		exec_eject,
		"/dev/rod0c"
	},
	{
		"format",
		"(physical format a media)",
		exec_format,
		"/dev/rod0.ctl"
	},
	{
		"verify",
		"(verify a media)",
		exec_verify,
		"/dev/rod0.ctl"
	},
	{
		NULL,
		NULL,
		NULL,
		NULL
	}
};

/*
 * global variables
 */

char *program = "odcontrol";
char *device;
int timeouts;
int verbose;
int have_message;
struct scsi_addr unit_addr;

/*
 * external reference
 */

extern char *scsi_sense_desc();

/*
 * SCSI utility routines
 *  ONLY for little endian machines.
 *  MSB at the lowest address, LSB at the highest.
 */

void
scsi_uto2b(val, bytes)
	u_int val;
	u_char *bytes;
{
	*bytes++ = (val & 0xff00) >> 8;
	*bytes = val & 0xff;
}

u_int
scsi_2btou(bytes)
	u_char *bytes;
{
	u_int rc;

        rc = (*bytes++ << 8);
        rc += *bytes;

        return rc;
}

void
scsi_uto4b(val, bytes)
	u_int val;
	u_char *bytes;
{
	*bytes++ = (val & 0xff000000) >> 24;
	*bytes++ = (val & 0xff0000) >> 16;
	*bytes++ = (val & 0xff00) >> 8;
	*bytes =    val & 0xff;
}

u_int
scsi_4btou(bytes)
	u_char *bytes;
{
	u_int rc;

	rc = (*bytes++ << 24);
	rc += (*bytes++ << 16);
	rc += (*bytes++ << 8);
	rc +=  *bytes;

	return rc;
}

/*
 * get tartget unit address
 */

int
scsi_identify(fd)
	int fd;
{
	int sts;

	sts = ioctl(fd, SCIOCIDENTIFY, &unit_addr);

	return sts;
}

/*
 * print SCSI error
 */

void
scsi_error(req)
	scsireq_t *req;
{
	struct sense_data *sd;
	int skey, asc, ascq;
	static char *retstat[] = {
		"Timeout", "Busy", "Sense Key", "Unknown"
	};
	static char *sense_keys[] = {
		"No Sense", "Recoverd Error", "Not Ready",
		"Medium Error", "Hardware Error", "Illegal Request",
		"Unit Attention", "Data Protect", "Blank Check",
		"Vendor Unique", "Copy Aborted", "Aborted Command",
		"Equal", "Volume Overflow", "Reserved"
	};

	if (have_message == 2) {
		fprintf(stderr, ", ");
	}
	if (have_message) {
		fprintf(stderr, "aborted\n");
	}
	fprintf(stderr, "%s: SCSI error: %s",
		program, retstat[req->retsts - 1]);
	if (req->retsts == SCCMD_SENSE) {
		sd = (struct sense_data *) req->sense;
		skey = sd->sense_key & SCSI_SENSE_KEY;
		fprintf(stderr, " - %s\n", sense_keys[skey]);
		if (sd->add_sense_len >= 5) {
			asc = sd->asc;
		} else {
			asc = 0;
		}
		if (sd->add_sense_len >= 6) {
			ascq = sd->ascq;
		} else {
			ascq = 0;
		}
		fprintf(stderr, "%s: [%02X-%02X] %s\n",
			program, asc, ascq, scsi_sense_desc(asc, ascq));
	} else {
		fprintf(stderr, "\n");
	}
}

/*
 * test unit ready
 */

int
test_unit_ready(fd)
	int fd;
{
	int sts;
	scsireq_t scsireq;
	struct test_unit_ready *tur;
	struct sense_data *sd;
	int skey;

	bzero(&scsireq, sizeof (scsireq_t));
	if (timeouts > 0) {
		scsireq.timeout = timeouts;
	} else {
		scsireq.timeout = DEFAULT_TIMEOUT;	/* 10 sec */
	}

	tur = (struct test_unit_ready *) scsireq.cmd;
	tur->op_code = SCSI_TEST_UNIT_READY;
	tur->lun = (unit_addr.lun << 5);

	scsireq.cmdlen = sizeof (struct test_unit_ready);
	scsireq.senselen = SENSEBUFLEN;

	sts = ioctl(fd, SCIOCCOMMAND, &scsireq);
	if (sts < 0) {
		perror(program);
		return ERR_SYSCALL;
	}
	if (scsireq.retsts != SCCMD_OK) {
		sd = (struct sense_data *) scsireq.sense;
		skey = sd->sense_key & SCSI_SENSE_KEY;
		if (scsireq.retsts != SCCMD_SENSE) {
			scsi_error(&scsireq);
			sts = ERR_SCSI;
		} else if (skey == SCSI_NO_SENSE || skey == SCSI_NOT_READY
			   || skey == SCSI_UNIT_ATTENTION) {
			/*
			 * "No Sense", "Not Ready" and "Unit Attention"
			 * are not really errors here.
			 */
			sts = ERR_NONE;
		} else {
			scsi_error(&scsireq);
			sts = ERR_SCSI;
		}
	}

	return sts;
}	

/*
 * get current cache control setting
 */

int
get_cache(fd, curval, len)
	int fd;
	u_char *curval;
	int len;
{
	int sts;
	scsireq_t scsireq;
	struct mode_sense *ms;

	bzero(&scsireq, sizeof (scsireq));
	bzero(curval, len);
	scsireq.flags = SCCMD_READ;
	if (timeouts > 0) {
		scsireq.timeout = timeouts;
	} else {
		scsireq.timeout = DEFAULT_TIMEOUT;	/* 10 sec */
	}

	ms = (struct mode_sense *) scsireq.cmd;
	ms->op_code = SCSI_MODE_SENSE;
	ms->lun_flag = (unit_addr.lun << 5);
	ms->page = 8;
	ms->alloc_len = len;

	scsireq.cmdlen = sizeof (struct mode_sense);
	scsireq.databuf = curval;
	scsireq.datalen = len;
	scsireq.senselen = SENSEBUFLEN;

	sts = ioctl(fd, SCIOCCOMMAND, &scsireq);
	if (sts < 0) {
		perror(program);
		return ERR_SYSCALL;
	}
	if (scsireq.retsts != SCCMD_OK) {
		scsi_error(&scsireq);
		sts = ERR_SCSI;
	}

	return sts;
}

/*
 * get changable cache control parameter mask
 */

int
get_cache_mask(fd, maskval, len)
	int fd;
	u_char *maskval;
	int len;
{
	int sts;
	scsireq_t scsireq;
	struct mode_sense *ms;

	bzero(&scsireq, sizeof (scsireq));
	bzero(maskval, len);
	scsireq.flags = SCCMD_READ;
	if (timeouts > 0) {
		scsireq.timeout = timeouts;
	} else {
		scsireq.timeout = DEFAULT_TIMEOUT;	/* 10 sec */
	}

	ms = (struct mode_sense *) scsireq.cmd;
	ms->op_code = SCSI_MODE_SENSE;
	ms->lun_flag = (unit_addr.lun << 5);
	ms->page = 0x40 | 8;
	ms->alloc_len = len;

	scsireq.cmdlen = sizeof (struct mode_sense);
	scsireq.databuf = maskval;
	scsireq.datalen = len;
	scsireq.senselen = SENSEBUFLEN;

	sts = ioctl(fd, SCIOCCOMMAND, &scsireq);
	if (sts < 0) {
		perror(program);
		return ERR_SYSCALL;
	}
	if (scsireq.retsts != SCCMD_OK) {
		scsi_error(&scsireq);
		sts = ERR_SCSI;
	}

	return sts;
}

/*
 * set cache control parameter
 */

int
set_cache(fd, newval, len)
	int fd;
	u_char *newval;
	int len;
{
	int sts;
	scsireq_t scsireq;
	struct mode_select *ms;

	bzero(&scsireq, sizeof (scsireq));
	scsireq.flags = SCCMD_WRITE;
	if (timeouts > 0) {
		scsireq.timeout = timeouts;
	} else {
		scsireq.timeout = DEFAULT_TIMEOUT;	/* 10 sec */
	}

	ms = (struct mode_select *) scsireq.cmd;
	ms->op_code = SCSI_MODE_SELECT;
	ms->lun_flag = (unit_addr.lun << 5) | 0x10;
	ms->plist_len= len;

	scsireq.cmdlen = sizeof (struct mode_select);
	scsireq.databuf = newval;
	scsireq.datalen = len;
	scsireq.senselen = SENSEBUFLEN;

	sts = ioctl(fd, SCIOCCOMMAND, &scsireq);
	if (sts < 0) {
		perror(program);
		return ERR_SYSCALL;
	}
	if (scsireq.retsts != SCCMD_OK) {
		scsi_error(&scsireq);
		sts = ERR_SCSI;
	}

	return sts;
}

/*
 * initiate format unit command
 */

int
format_unit(fd)
	int fd;
{
	int sts;
	scsireq_t scsireq;
	struct format_unit *fu;

	bzero(&scsireq, sizeof (scsireq_t));
	if (timeouts > 0) {
		scsireq.timeout = timeouts;
	} else {
		scsireq.timeout = DEFAULT_TIMEOUT_LONG;	/* 3600 sec */
	}

	fu = (struct format_unit *) scsireq.cmd;
	fu->op_code = SCSI_FORMAT_UNIT;
	fu->lun_flag = (unit_addr.lun << 5);
	fu->interleave_1 = 0;	/* default interleave */
	fu->interleave_0 = 0;

	scsireq.cmdlen = sizeof (struct format_unit);
	scsireq.databuf = NULL;
	scsireq.datalen = 0;
	scsireq.senselen = SENSEBUFLEN;

	sts = ioctl(fd, SCIOCCOMMAND, &scsireq);
	if (sts < 0) {
		perror(program);
		return ERR_SYSCALL;
	}
	if (scsireq.retsts != SCCMD_OK) {
		scsi_error(&scsireq);
		sts = ERR_SCSI;
	}

	return sts;
}

/*
 * get unit capacity
 */

int
read_capacity(fd, max)
	int fd;
	u_int *max;
{
	int sts;
	scsireq_t scsireq;
	struct read_capacity *rc;
	struct read_capacity_data rdcap_data;

	bzero(&scsireq, sizeof (scsireq));
	bzero(&rdcap_data, sizeof (rdcap_data));
	scsireq.flags = SCCMD_READ;
	if (timeouts > 0) {
		scsireq.timeout = timeouts;
	} else {
		scsireq.timeout = DEFAULT_TIMEOUT;	/* 10 sec */
	}

	rc = (struct read_capacity *) &scsireq.cmd;
	rc->op_code = SCSI_READ_CAPACITY;
	rc->lun_flag = (unit_addr.lun << 5);

	scsireq.cmdlen = sizeof (struct read_capacity);
	scsireq.databuf = (u_char *) &rdcap_data;
	scsireq.datalen = sizeof (rdcap_data);
	scsireq.senselen = SENSEBUFLEN;

	sts = ioctl(fd, SCIOCCOMMAND, &scsireq);
	if (sts < 0) {
		perror(program);
		return ERR_SYSCALL;
	}
	if (scsireq.retsts != SCCMD_OK) {
		scsi_error(&scsireq);
		return ERR_SCSI;
	}

	*max = scsi_4btou(&rdcap_data.block_addr_3) + 1;

	return ERR_NONE;
}

/*
 * verify media blocks
 */

int
verify_media(fd, addr, len, bad)
	int fd;
	u_int addr;
	u_int len;
	u_int *bad;
{
	int sts;
	scsireq_t scsireq;
	struct verify *vc;
	struct sense_data *sense;

	bzero(&scsireq, sizeof (scsireq));
	if (timeouts > 0) {
		scsireq.timeout = timeouts;
	} else {
		scsireq.timeout = DEFAULT_TIMEOUT;	/* 10 sec */
	}

	vc = (struct verify *) &scsireq.cmd;
	vc->op_code = SCSI_VERIFY;
	vc->lun_flag = (unit_addr.lun << 5);
	scsi_uto4b(addr, &vc->block_addr_3);
	scsi_uto2b(len, &vc->block_len_1);

	scsireq.cmdlen = sizeof (struct verify);
	scsireq.databuf = NULL;
	scsireq.datalen = 0;
	scsireq.senselen = SENSEBUFLEN;

	sts = ioctl(fd, SCIOCCOMMAND, &scsireq);
	if (sts < 0) {
		perror(program);
		return ERR_SYSCALL;
	}
	if (scsireq.retsts != SCCMD_OK) {
		sense = (struct sense_data *) scsireq.sense;
		if (scsireq.retsts == SCCMD_SENSE
		    && (sense->sense_key & SCSI_SENSE_KEY)
		       == SCSI_MEDIUM_ERROR) {
			*bad = scsi_4btou(&sense->info_3);
			sts = ERR_BADBLOCK;
		} else {
			scsi_error(&scsireq);
			sts = ERR_SCSI;
		}
	}

	return sts;
}

/*
 * reassign bad block
 */

int
reassign_block(fd, bad)
	int fd;
	u_int bad;
{
	int sts;
	scsireq_t scsireq;
	struct reassign_blocks *rb;
	struct defect_list defect;

	bzero(&scsireq, sizeof (scsireq));
	scsireq.flags = SCCMD_WRITE;
	if (timeouts > 0) {
		scsireq.timeout = timeouts;
	} else {
		scsireq.timeout = DEFAULT_TIMEOUT;	/* 10 sec */
	}

	rb = (struct reassign_blocks *) scsireq.cmd;
	rb->op_code = SCSI_REASSIGN_BLOCKS;
	rb->lun = (unit_addr.lun << 5);

	bzero(&defect, sizeof (defect));
	scsi_uto2b(4, &defect.list_len_1);
	scsi_uto4b(bad, &defect.block_addr_3);

	scsireq.cmdlen = sizeof (struct reassign_blocks);
	scsireq.databuf = (caddr_t) &defect;
	scsireq.datalen = sizeof (defect);
	scsireq.senselen = SENSEBUFLEN;

	sts = ioctl(fd, SCIOCCOMMAND, &scsireq);
	if (sts < 0) {
		perror(program);
		return ERR_SYSCALL;
	}
	if (scsireq.retsts != SCCMD_OK) {
		scsi_error(&scsireq);
		sts = ERR_SCSI;
	}

	return sts;
}

/*
 * current drive cache mode control
 */

int
exec_cache(argc, argv)
	int argc;
	char **argv;
{
	int fd;
	u_char data[255];
	struct cache_ctl page8, *ctl;
	struct mode_parm_hdr pheader, *mph;
	int curval, mask, newval;
	int sts;
	char *arg, **p;
	int n;

	fd = open(device, O_RDWR);
	if (fd < 0) {
		perror(program);
		return ERR_SYSCALL;
	}

	/* get tartget address */
	sts = scsi_identify(fd);
	if (sts < 0) {
		perror(program);
		close(fd);
		return ERR_SYSCALL;
	}

	/* clear "Unit Attention" */
	sts = test_unit_ready(fd);
	if (sts != 0) {
		close(fd);
		return sts;
	}

	/* get current setting */
	sts = get_cache(fd, data, 255);
	if (sts != 0) {
		close(fd);
		return sts;
	}

	mph = (struct mode_parm_hdr *) data;
	bcopy(mph, &pheader, sizeof (struct mode_parm_hdr));
	ctl = (struct cache_ctl *)
	  (data + sizeof (struct mode_parm_hdr) + mph->blk_desc_len);
	bcopy(ctl, &page8, sizeof (struct cache_ctl));
	curval = ctl->cache_mode;

	if (argc != 0) {	/* set cache mode */
		/* find argument */
		arg = *argv;
		for (n = 0, p = cache_args; *p != NULL; ++p) {
			if (strcasecmp(arg, *p) == 0) {
				break;
			}
			n++;
		}
		if (*p == NULL) {
			fprintf(stderr, "%s: unknown argument: %s\n",
				program, arg);
			close(fd);
			return ERR_COMMAND;
		}

		/* get new setting value */
		newval = 0;
		switch (n) {
		      case C_CACHE_OFF:	/* off */
			newval = SCSI_MODE_PAGE8_RCD;
			newval &= ~SCSI_MODE_PAGE8_WCE;
			break;
		      case C_CACHE_WR:	/* write enable */
			newval = SCSI_MODE_PAGE8_WCE | SCSI_MODE_PAGE8_RCD;
			break;
		      case C_CACHE_RD:	/* read enable */
			newval &= ~SCSI_MODE_PAGE8_RCD;
			newval &= ~SCSI_MODE_PAGE8_WCE;
			break;
		      case C_CACHE_ON:	/* on */
			newval &= ~SCSI_MODE_PAGE8_RCD;
			newval |= SCSI_MODE_PAGE8_WCE;
			break;
		}

		/* get changable mask data */
		sts = get_cache_mask(fd, data, 255);
		if (sts != 0) {
			close(fd);
			return sts;
		}
		mph = (struct mode_parm_hdr *) data;
		ctl = (struct cache_ctl *)
		  (data + sizeof (struct mode_parm_hdr) + mph->blk_desc_len);
		mask = ctl->cache_mode;

		/* chage only changable data */
		if (!(mask & SCSI_MODE_PAGE8_RCD)) {
			newval &= ~SCSI_MODE_PAGE8_RCD;
			newval |= (curval & SCSI_MODE_PAGE8_RCD);
		}
		if (!(mask & SCSI_MODE_PAGE8_WCE)) {
			newval &= ~SCSI_MODE_PAGE8_WCE;
			newval |= (curval & SCSI_MODE_PAGE8_WCE);
		}

		curval &= ~(SCSI_MODE_PAGE8_WCE | SCSI_MODE_PAGE8_RCD);
		newval |= curval;

		/* set mode parameter header */
		mph = (struct mode_parm_hdr *) data;
		bcopy(&pheader, data, sizeof (struct mode_parm_hdr));
		mph->mode_parm_len = 0;
		mph->blk_desc_len = 0;

		/* set cache control parameter (page 8) */
		ctl = (struct cache_ctl *)
		  (data + sizeof (struct mode_parm_hdr));
		bcopy(&page8, ctl, sizeof (struct cache_ctl));
		ctl->cache_mode = newval;
		ctl->page &= 0x7f;		/* clear PS bit */

		/* set cache control parameter (page 8) */
		n = sizeof (struct mode_parm_hdr) + sizeof (struct cache_ctl);
		sts = set_cache(fd, data, n);
		if (sts != 0) {
			close(fd);
			return sts;
		}

		/* re-read cache control parameter (page 8) */
		sts = get_cache(fd, data, 255);
		if (sts != 0) {
			close(fd);
			return sts;
		}
		mph = (struct mode_parm_hdr *) data;
		ctl = (struct cache_ctl *)
		  (data + sizeof (struct mode_parm_hdr) + mph->blk_desc_len);
		curval = ctl->cache_mode;
	}
	
	/* print current setting */
	printf("read cache: %s\n",
	       (curval & SCSI_MODE_PAGE8_RCD) ? "off" : "on");
	printf("write cache: %s\n",
	       (curval & SCSI_MODE_PAGE8_WCE) ? "on" : "off");

	close(fd);

	return sts;
}

/*
 * eject loaded media
 */

int
exec_eject(argc, argv)
	int argc;
	char **argv;
{
	int fd;
	int sts;

	fd = open(device, O_RDONLY);
	if (fd < 0) {
		if (errno == ENXIO) {
			/* no media */
			return ERR_NONE;
		}
		return ERR_SYSCALL;
	}
	sts = ioctl(fd, CDIOCALLOW);
	if (sts >= 0) {
		ioctl(fd, CDIOCEJECT);
		sts = ERR_NONE;
	} else {
		sts = ERR_SYSCALL;
	}
	close(fd);

	return ERR_NONE;
}

/*
 * format media
 */

int
exec_format(argc, argv)
	int argc;
	char **argv;
{
	int fd;
	int sts;

	fd = open(device, O_RDWR);
	if (fd < 0) {
		perror(program);
		return ERR_SYSCALL;
	}

	/* get tartget address */
	sts = scsi_identify(fd);
	if (sts < 0) {
		perror(program);
		close(fd);
		return ERR_SYSCALL;
	}

	/* clear "Unit Attention" */
	sts = test_unit_ready(fd);
	if (sts != 0) {
		close(fd);
		return sts;
	}

	printf("formatting a media (scbus%d:target%d:lun%d) ... ",
	       unit_addr.scbus, unit_addr.target, unit_addr.lun);
	fflush(stdout);
	have_message = 1;

	sts = format_unit(fd);
	if (sts == 0) {
		printf("done\n");
	}

	close(fd);

	return sts;
}

/*
 * verify media
 */

int
exec_verify(argc, argv)
	int argc;
	char **argv;
{
	int fd;
	int sts;
	u_int max, addr, len, bad;
	char *mesg = "\b\b\b\b\b\b\b\b\b%3u%% done";

	fd = open(device, O_RDWR);
	if (fd < 0) {
		perror(program);
		return ERR_SYSCALL;
	}

	/* get tartget address */
	sts = scsi_identify(fd);
	if (sts < 0) {
		perror(program);
		close(fd);
		return ERR_SYSCALL;
	}

	/* clear "Unit Attention" */
	sts = test_unit_ready(fd);
	if (sts != 0) {
		close(fd);
		return sts;
	}

	sts = read_capacity(fd, &max);
	if (sts != 0) {
		close(fd);
		return sts;
	}

	printf("verifying a media (scbus%d:target%d:lun%d) ... %3u%% done",
	       unit_addr.scbus, unit_addr.target, unit_addr.lun, 0);
	fflush(stdout);
	have_message = 2;

	for (addr = 0; addr < max; addr += len) {
		len = max - addr;
		if (len > MAX_VERIFY_LENGTH) {
			len = MAX_VERIFY_LENGTH;
		}
		sts = verify_media(fd, addr, len, &bad);
		if (sts < 0) {
			sts = reassign_block(fd, bad);
			if (sts != 0) {
				break;
			}
			len = bad - addr;
		} else if (sts != 0) {
			break;
		}
		printf(mesg, (addr * 100) / max);
		fflush(stdout);
	}

	if (sts == 0) {
		printf(mesg, (addr * 100) / max);
		printf("\n");
		sleep(1);
	}
	close(fd);

	return sts;
}

/*
 * print usage
 */

void
usage()
{
	struct cmd_tab *p;

	fprintf(stderr,
		"usage: %s [-f device] [-t timeout] command [arg]\n",
		program);
	fprintf(stderr, "\tcommands:\n");
	for (p = commands; p->c_name != NULL; ++p) {
		fprintf(stderr, "\t  %s %s\n", p->c_name, p->help);
	}
	exit(ERR_COMMAND);
}

/*
 *  main
 */

void
main(argc, argv)
	int argc;
	char **argv;
{
	int ch;
	struct cmd_tab *cmd;
	int sts;
	struct stat sb;

	device = NULL;
	have_message = verbose = 0;
	timeouts = 0;

	while ((ch = getopt(argc, argv, "f:t:v?")) != EOF) {
		switch (ch) {
		case 'f':
			device = optarg;
			break;
		case 't':
			timeouts = atoi(optarg);
			if (timeouts < 1) {
				usage();
			}
			timeouts *= 1000;
			break;
		case 'v':
			verbose = 1;
			break;
		case '?':
		default:
			usage();
		}
	}
	argc -= optind;
	argv += optind;

	if (argc == 0) {
		usage();
	}

	for (cmd = commands; cmd->c_name != NULL; ++cmd) {
		if (strcasecmp(cmd->c_name, *argv) == 0) {
			argc--;
			argv++;
			if (device == NULL) {
				device = cmd->def_dev;
			}
			sts = stat(device, &sb);
			if (sts < 0) {
				perror(program);
				exit(ERR_COMMAND);
			}
			sts = (cmd->func)(argc, argv);
			break;
		}
	}
	if (cmd->c_name == NULL) {
		fprintf(stderr, "%s: unknown command: %s\n", program, *argv);
		usage();
	}

	exit(sts);
}
